#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#include <time.h>

#include "structures.h"

solver* my_m;                   // needed by signal handler only
printmode print_mode = RESULT;
bool verbose = false;
double time_out;                //initialized in parse_options

// interrupt
void SIGINT_handler(int signum) 
{
  my_m->print_progress();
  my_m->print_progress_footer();  
  rprintf("\n");
  rprintf("\n");
  rprintf("INTERRUPTED\n");
  my_m->print_stats();  
  rprintf("Time used: %fs",get_cpu_time());
  exit(0); 
}

// seg fault
void SIGSEGV_handler(int signum) 
{
  my_m->print_progress();
  my_m->print_progress_footer();  
  rprintf("\n");
  rprintf("\n");
  rprintf("SEGMENTATION FAULT\n");
  printf("s UNKNOWN\n");
  rprintf("\n");
  my_m->print_stats();
  exit(3); 
}

void print_options()
{
  printf("STLSat 1.00 options:\n"
         "\n"
         " -q \tquiet. Do not print out the answer line. Suppress -s.\n"
         " -r \trestarts off.  Do not restart the solver.\n"
         " -s \tsolution. Print out solution if one is found.\n"
         " -t n \ttime-out. Stop and return UNKNOWN after <n> seconds.\n"
         " -v \tverbose. Print out useful information during execution.\n"
         "\n"
         "Example:\n"
         "\t./stlsat problem.cnf -s -t 100 -v\n");
}

void print_usage()
{
  printf("Usage: stlsat <cnf-file-name> [options]\n"
         "Solve the SAT problem specified in <cnf-file-name>.\n"
         "Example: stlsat sat-problem.cnf\n");
  print_options();
}

void parse_options(size_t num, char** argv)
{
  for (size_t i = 1 ; i < num ; ++i) 
    if (argv[i][0] == '-') switch (argv[i][1]) {
      case 't': 
        if (i == num - 1) {
          printf("Expecting a time-out argument following -t.\n");
          exit(0);
        }
        time_out = atof(argv[++i]);
        break;
      case 'q': print_mode = SILENT; break;
      case 'r': g_restarts_on = false; break;
      case 's': if (print_mode == SILENT) print_mode = SOLUTION; break;
      case 'v': verbose = true; break;
      case 'h': print_usage(); exit(0);
      }

  if (time_out > 0 && print_mode != SILENT)
    rprintf("Time out set to %.4f seconds\n",time_out);
}

#define STLSAT_VERSION 2.01

void show_result(const string &str)
{
  rprintf("\n");
  rprintf("\n");
  printf("s %s\n",str.c_str());
}

int main(int argc, char *argv[])
{
  signal(SIGINT,SIGINT_handler);
  signal(SIGSEGV,SIGSEGV_handler);

  if (argc < 2) { print_usage(); return 0; }
  
  time_out = -1.0;
  parse_options(argc,argv);

  if(print_mode != SILENT) rprintf("STLSat version %0.2f\n",STLSAT_VERSION);
  fflush(stdout);

  vector<set<UPTestingLit> > clauses = read_cnf(argc,argv);
  bool ok;
  solver s(clauses,ok);
  
  my_m = &s;                    // needed by signal handlers
  rprintf("solving instance: %s\n",argv[1]);

  if (ok) ok = s.finish_initialization();

  s.print_mode() = print_mode;
  s.verbose() = verbose;

  Status result = ok? SATISFIABLE : UNSATISFIABLE;

  if (ok && !s.empty()) {       // if trivial, you're done!
    s.print_progress_header();

    t_start_time = get_cpu_time();
    result = s.solve(); 
    t_solution_time = get_cpu_time() - t_start_time;
    switch (result) {
    case UNSATISFIABLE: t_result = UNSAT; break;
    case SATISFIABLE: t_result = SAT; break;
    case UNKNOWN: t_result = TIME_OUT; break;
    case SOLVING: t_result = SAMPLE_FINISHED; break;
    }
    output_solver_stats();
    output_result(t_result);
    if (result != SATISFIABLE && result != UNSATISFIABLE) {
      s.print_progress();
      s.print_progress_footer();
    }
      
    s.backtrack(1);             // clean up
  }

  switch (result) {
  case SATISFIABLE: 
    if (print_mode != SILENT) show_result("SATISFIABLE");
    break;
  case UNSATISFIABLE:
    if (print_mode != SILENT) show_result("UNSATISFIABLE");
    break;
  default: show_result("UNKNOWN");
  }

  s.print_stats();
  rprintf("Running time: %.5f seconds\n",get_cpu_time());

  switch (result) {             // status as reqd by SAT competitions
  case SATISFIABLE: exit(10);
  case UNSATISFIABLE: exit(20);
  default: exit(0);
  }
}
