#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <math.h>

#include "structures.h"

extern double time_out;  //time out limit

// Luby function.  n is index into sequence; base is first element.
int get_luby(int n, int base)
{
  if (n == 1) return base;

  int k = -1;                    // floor (log2(n + 1))
  for (int m = n + 1 ; m ; m >>= 1) ++k;
  if (1 << k == n + 1) return base * (1 << (k - 1));
  return get_luby(n + 1 - (1 << k),base);
}

// As we build up a conflict clause, we accumulate literals from
// reasons.  Now, if we have a new literal, we might want to add it to
// the clause, or we might not need to because this literal will
// itself be resolved on at a later step.  This function figures that
// out.  The arguments are the literal in question, a vector<bool>
// indicating whether or not each assertion level is part of the
// resolution, and then a vector<bool> indicating which variables have
// already been part of the resolution.

inline bool solver::removable(literal l, vector<bool> &levels, 
                              vector<bool> &seen) const
{  
  static vector<literal> stack; // literals you're trying to figure out
  static vector<atom> save;     // atoms you *think* you've seen
  stack.clear();
  save.clear();
  
  stack.push_back(l);           // initially, just figuring out this one
  
  while (!stack.empty()) {
    const clause &rsn = (*this)[reason(get_atom(stack.back()))];
    stack.pop_back();
    for (size_t i = 1 ; i < rsn.size() ; ++i) { // each literal in the reason
      literal lit = rsn[i];
      atom a = get_atom(lit);
      size_t lvl = level(a);
      if (seen[a]) continue;    // already done, so not a problem
      if (lvl == 1) continue;   // can't back up further

      if (!reason(a) ||         // branch choice -- not removable!
          !levels[lvl]) {       // this level not involved -- not removable!
        for (size_t j = 0 ; j < save.size() ; ++j) seen[save[j]] = false;
        return false;           // reset seen to old value and give up
      }

      seen[a] = true;           // looks good; now have to consider antecedent
      stack.push_back(lit);
      save.push_back(a);
    }
  }
  return true;                  // success!
}

// Derive a conflict.  Keep resolving literals at the given level
// until there is only one left.

clause solver::derive_conflict(clause_id conf_id, size_t clevel, size_t& alevel)
{
  clause ans;
  static vector<bool> seen;
  seen.resize(variables.size());
  seen.assign(variables.size(),false);
  int num_lits_at_clevel = 0;
  literal next_lit = 0;         // 0 means not yet started
  vector<literal>::reverse_iterator next_on_stack = assign.rbegin();
  
  if (clevel <= 1) {            // conflict at top of tree -- fatal
    alevel = 0;
    return ans;
  }

  alevel = 1;                   // how far to backtrack
  ans.push_back(0);             // asserted literal will go here

  do {
    assert(conf_id);
    const clause &conf = (*this)[conf_id];
    for (size_t i = (next_lit? 1 : 0) ; i < conf.size() ; ++i) {
      // skip first lit after first clause; it's what we're resolving
      literal lit = conf[i];
      atom a = get_atom(lit);
      size_t lvl = level(a);

      if (!seen[a] && (lvl > 1)) { // unless already done, or at top level ...
        seen[a] = true;            // it's done now
        increment_literal_score(lit);
        if (lvl == clevel) ++num_lits_at_clevel; // either another lit this level
        else ans.push_back(lit);            // or it's part of the resolvent
      }
    }

    // get next seen variable (i.e., next variable known to be in conflict)
    while (!seen[get_atom(*next_on_stack)]) ++next_on_stack;
    next_lit = *next_on_stack; 
    atom a = get_atom(next_lit);
    seen[a] = false;         // mark as unseen to exclude it from learned clause
    --num_lits_at_clevel;    // just handled one!

    if ((conf_id = reason(a)) > 0) 
      increment_clause_score(conf_id); // increment learned clause only
  } while (num_lits_at_clevel > 0);

  ans[0] = negation(next_lit); // asserted literal
  seen[get_atom(next_lit)] = true;

  // now simplify derived clause
  static vector<bool> levels;
  levels.resize(decision_lit.size());
  levels.assign(decision_lit.size(),false);
  for (size_t i = 1; i < ans.size() ; i++) 
    levels[level(get_atom(ans[i]))] = true; // levels with a conflict

  // now retain each literal if it's a branch point or not removable
  size_t j = 1;
  for (size_t i = 1 ; i < ans.size() ; i++){
    atom a = get_atom(ans[i]);
    if(!reason(a) || !removable(ans[i],levels,seen)) {
      alevel = max(level(a),alevel);
      ans[j++] = ans[i];
    }
  }

  ans.resize(j);
  return ans;
}

// Flip the status of the progress saving stuff, incrementing the
// various counters.

void saving::flip_status()
{
  if (save_progress) {
    save_progress = false;
    off_th += off_th_inc;
    next_switch += on_th;
  }
  else {
    save_progress = true;
    on_th += on_th_inc;
    next_switch += off_th;
  }
}

void solver::backtrack(size_t dest)
{
  if (decision_lit.size() < dest) return;
  // may happen if restart while at level 1 (top).  Don't erase
  // assignments from previous levels.

  for (vector<literal>::reverse_iterator stack = assign.rbegin() ;
       stack < assign.rend() - decision_lit[dest - 1] ; ++stack) {
    literal l = *stack;
    atom a = get_atom(l);
    
    if (save.save()) saved(a) = l; // save progress if requested
    
    reason(a) = 0;
    level(a)  = 0;
    status(a) = 0;
    var_order_heap.insert_if_missing(a); // back in heap
  }
  
  assign.resize(decision_lit[dest - 1]);
  decision_lit.resize(dest - 1);
}

// Handle initial unit literal queue; return conflict if found
clause_id solver::process_unit_literal_queue()
{
  if (units.empty()) return 0;

  simplify_orig_kb = true;
  simplify_learned_kb = true;
  atom a = units.front();
  return bcp(variables[a].status,false);    
}

// Unit propagate, starting with given literal.  Return pointer to
// conflict or NULL if all OK.  Second argument indicates is true if
// the assignment needs to be added to the stack
clause_id solver::bcp(literal l, bool push)
{
  size_t slevel = level(get_atom(l)); // level of assignment
  units.insert_if_missing(get_atom(l)); // add to queue
  if (push) assign.push_back(l);
  status(get_atom(l)) = l;

  while (!units.empty()) {
    atom a = units.front();
    ++queue_pops;
    literal lit = status(a);
    literal neg_lit = negation(lit);
    int wi = index(neg_lit);    // index for watching
    vector<clause_id>::iterator pointer = watched[wi].begin(); 
                                // reassemble list here
    for (vector<clause_id>::iterator watcher = pointer ; 
         watcher < watched[wi].end() ; ++watcher) { // iterate over watched
      ++clauses_touched;
      clause &clause = (*this)[*watcher];
      if (clause[1] != neg_lit) swap(clause[0],clause[1]);
      // clause[1] is false; clause[0] is where the action is

      if (is_set(clause[0])) { *pointer++ = *watcher; continue; }
      // do nothing if clause sat

      size_t k;
      for (k = 2 ; k < clause.size(); ++k) { // search for new unassigned
        ++literals_touched;
        if (is_false(clause[k])) continue;
        swap(clause[1],clause[k]); // found it -- move to second lit
        lit_watches_clause(clause[1],clause);
        break;
      }
      if (k < clause.size()) continue; // done with clause

      // couldn't find unassigned literal; clause is unit or unsat
      if (is_false(clause[0])) { // clause unsat
        units.clear();
	while (watcher < watched[wi].end()) *pointer++ = *watcher++;
        watched[wi].erase(pointer,watched[wi].end());
        return clause.id;
      }
      else if (is_free(clause[0])) {
        //this literal is implied for the first time
        atom a = get_atom(clause[0]);
        reason(a) = clause.id;	  
        status(a) = clause[0];
        level(a) = slevel;
        units.insert_if_missing(a); // add to queue
        assign.push_back(clause[0]);
	*pointer++ = *watcher;
      }
    }
    watched[wi].erase(pointer,watched[wi].end());
  }
  return 0;
}

// Add a new conflict clause and unit propagate, returning clause if conflict
clause_id solver::assert_conflict(clause &conf, size_t lev)
{
  if (conf.size() > 1) add_conflict(conf,lev);    // unit handled elsewhere

  atom a = get_atom(conf[0]);
  level(a) = lev;
  reason(a) = (conf.size() > 1)? conf.id : 0;
  
  if (conf.size() == 1) simplify_orig_kb = simplify_learned_kb = true;

  score_inc /= SCORE_INC_FACTOR;
  clause_score_inc /= CLAUSE_SCORE_INC_FACTOR;
  
  return bcp(conf[0]);
}

// Set a literal.  Add to assignment and unit propagate.
clause_id solver::set_decision(literal l) 
{
  decision_lit.push_back(assign.size());  
  ++decisions;  
  
  level(get_atom(l)) = decision_lit.size();
  reason(get_atom(l)) = 0;
  return bcp(l);
}

atom solver::first_unvalued() const
{
  ++number_branch_decisions;
  for (size_t i = 1 ; i < variables.size() ; ++i) 
    if (status(i) == 0) return i;
  return 0;                     // no free variables, so SAT!
}

atom solver::select_variable()
{
  ++number_branch_decisions;
  while (!var_order_heap.empty()) {
    atom next = var_order_heap.front(); 
    if (level(next) == 0) return next;
  }
  return 0;                     // no free variables, so SAT!
}

bool timed_out_global()
{
  return (t_time_out > 0 && t_time_out < (get_cpu_time() - t_start_time));
}

bool solver::timed_out() const
{
  return decisions % 2000 == 0 && time_out > 0 && get_cpu_time() > time_out;
}

Status solver::terminate(Status r)
{
  backtrack();
  return r;
}

void solver::do_restart()
{	
  print_progress();
  backtrack();
  r.next = conflicts + get_luby(++r.number + 1,r.luby_unit);
  if (g_forget_clauses_on) simplify(false);
}

void solver::do_solved()
{      
  print_progress();
  print_progress_footer();
  if (print_mode() == SOLUTION) print_solution();	
  backtrack();
}

Status solver::solve()
{
  start_time = clock();
  previous = statistics();
  Status result = SOLVING;

  // at each iteration in the following, unit propagation is complete
  // and there is no conflict.  We restart if need be, manage the kb
  // if need be, then pick a decision variable and branch.

  for ( ;; ) {                  // each iteration is a decision
    if (result == UNSATISFIABLE) return result; // UNSAT -- return it

    if (timed_out_global()) return terminate(UNKNOWN);
    if (sample_size && number_branch_decisions >= sample_size) 
      return terminate(SOLVING);
    if (timed_out()) return terminate(UNKNOWN);
	
    if (g_restarts_on && conflicts >= r.next) do_restart();

    if (g_forget_clauses_on) {
      if (simplify_learned_kb && decision_lit.size() == 1 &&
          conflicts >= num_conflicts_for_next_simplify)
        simplify();
      if (learned_clauses.size() >= max_learned_clauses + assign.size()) 
        reduce();
    }

    atom decision = 
      g_branching_heuristic_on? select_variable() : first_unvalued();

    if (!decision) { do_solved(); return SATISFIABLE; }

    literal lit;
    if (saved(decision) && save.save()) lit = saved(decision);
    else lit = nlit(decision);

    // in following, set_decision and assert_conflict return non-NULL if conflict
    clause c;
    size_t alevel;
    for (clause_id conf = set_decision(lit) ; conf ; 
         conf = assert_conflict(c,alevel)) {
      save.check_status(conflicts++);
      c = derive_conflict(conf,decision_lit.size(),alevel);
      // get conflict clause
      if (alevel == 0) return UNSATISFIABLE;      
      backtrack(alevel+1);
    }      
  }
  return result;
}
