// THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH
// (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL
// WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY
// WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
// RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR
// NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE
// FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting
// the generality of the foregoing, neither the author, nor JAIR, nor
// JAIR's publishers and distributors, warrant that the Source Code
// will be error-free, will operate without interruption, or will meet
// the needs of the user.

#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <math.h>

#include "structures.h"

// queue up a given literal for assignment
void solver::enqueue(literal l)
{ 
  atom a = get_atom(l);
  status(a) = l;                // atom now has given value
  level(a) = 1;                 // enqueue used at root of tree only
  reason(a) = 0;                // no reason needed
  units.insert_if_missing(a);   // new unit
  assign.push_back(l);          // push onto current assignment
}

// add a solver's _conflict_ clause to the database
void solver::add_conflict(clause &conf, size_t lev)
{
  bool found = false;
  
  // ensure that the conflict variable is second
  for (size_t i = 0 ; i < conf.size() ; ++i) {
    literal l = conf[i];
    increment_literal_score(l);
    if (!found && level(get_atom(l)) == lev) {
      swap(conf[1],conf[i]);
      found = true;
    }
  }
  conf.id = 1 + learned_clauses.size(); // it will be a new clause
  initialize_watchers(conf);
  
  learned_clauses.push_back(conf);
  learned_scores.push_back(0);
  
  ++clauses;                    // new clause learned
  learned_lits += conf.size();
  
  increment_clause_score(conf.id);
}

void solver::remove_clause(const clause &c)
{
  atom a0 = get_atom(c[0]);
  atom a1 = get_atom(c[1]);
  
  if (reason(a0) == c.id) reason(a0) = 0;
  else if (reason(a1) == c.id) reason(a1) = 0;

  lit_unwatches_clause(c[0],c);
  lit_unwatches_clause(c[1],c);
}

bool solver::is_satisfied(const clause &c) const
{
  for (size_t i = 0 ; i < c.size() ; ++i) 
    if (is_set(c[i]) > 0) return true; // sat
  return false;
}

// simplify learned clauses by removing those that are satisfied at
// the top level
void solver::simplify(bool learned)
{
  if (learned? !simplify_learned_kb : !simplify_orig_kb) return;
  size_t j = 0;                 // build up database of unsat clauses
  vector<clause> &clauses = learned? learned_clauses : original_clauses;
  int64 &ctr = learned? learned_lits : orig_lits;

  for (size_t i = 0 ; i < clauses.size() ; ++i)
    if (is_satisfied(clauses[i])) {
      ctr -= clauses[i].size();
      remove_clause(clauses[i]);
    }
    else {
      if (learned) learned_scores[j] = learned_scores[i];
      clauses[j++] = clauses[i];
    }

  clauses.resize(j);
  if (learned) learned_scores.resize(j);
  correct_watched_lists(!learned);

  if (learned) {
    kb_simplification_count++;
    num_conflicts_for_next_simplify = conflicts;
    simplify_learned_kb = false;
  }
  else simplify_orig_kb = false;
}

bool solver::is_reason(const clause &c) const // return true if c is a reason
{
  return c.size() <= 1 ||
    reason(get_atom(c[0])) == c.id || reason(get_atom(c[1])) == c.id;
}

// tiny structure used sort clauses based on their score.
struct sortable {
  size_t id;
  const solver *s;
  bool operator<(const sortable &other) const 
  { return s->learned_scores[id] < s->learned_scores[other.id]; }
};

void solver::reduce()
{
  if (learned_clauses.empty()) return;

  // sort learned clauses
  vector<sortable> tosort(learned_clauses.size());
  for (size_t i = 0 ; i < learned_clauses.size() ; ++i) {
    tosort[i].id = i;
    tosort[i].s = this;
  }
  sort(tosort.begin(),tosort.end());
  vector<double> scores = learned_scores;
  vector<clause> clauses = learned_clauses;
  for (size_t i = 0 ; i < learned_clauses.size() ; ++i) {
    learned_clauses[i] = clauses[tosort[i].id];
    learned_scores[i] = scores[tosort[i].id];
  }

  size_t j = 0;

  size_t bad_half = learned_clauses.size() / 2;
  double lim = (double)(clause_score_inc)/learned_clauses.size();
  for (size_t i = 0 ; i < learned_clauses.size() ; ++i) {
    const clause &cur = learned_clauses[i];
    if (!is_reason(cur) && cur.size() > 2 && 
        (i < bad_half || learned_scores[i] < lim)) {
      learned_lits -= cur.size();
      remove_clause(cur);
    }
    else {      
      learned_scores[j] = learned_scores[i];
      learned_clauses[j++] = learned_clauses[i];
    }
  }

  learned_clauses.resize(j);
  learned_scores.resize(j);
  correct_watched_lists(false);
  kb_reduction_count++;
  max_learned_clauses *= LEARNED_DB_MULTIPLIER;
}

static void adjust(clause_id &loc, bool orig, vector<clause_id> &renumber)
{
  if (orig == (loc < 0))
    if (loc < 0) loc = -1 - renumber[-1 - loc];
    else loc = renumber[loc];
}

void solver::correct_watched_lists(bool orig)
{
  vector<clause> &clauses = orig? original_clauses : learned_clauses;
  size_t s = clauses.size();
  vector<clause_id> renumber(s);
  for (size_t i = 0 ; i < clauses.size() ; ++i)
    if (orig) {
      if (-clauses[i].id - 1 >= (int) renumber.size())
        renumber.resize(-clauses[i].id);
      renumber[-clauses[i].id - 1] = i;
      clauses[i].id = -i - 1;
    }
    else {
      if (clauses[i].id >= (int) renumber.size())
        renumber.resize(clauses[i].id + 1);
      renumber[clauses[i].id] = i + 1;
      clauses[i].id = i + 1;
    }
  for (size_t i = 0 ; i < watched.size() ; ++i)
    for (size_t j = 0 ; j < watched[i].size() ; ++j)
      adjust(watched[i][j],orig,renumber);
  for (size_t i = 0 ; i < variables.size() ; ++i)
    if (reason(i)) adjust(reason(i),orig,renumber);
}
