// THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH
// (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL
// WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY
// WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
// RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR
// NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE
// FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting
// the generality of the foregoing, neither the author, nor JAIR, nor
// JAIR's publishers and distributors, warrant that the Source Code
// will be error-free, will operate without interruption, or will meet
// the needs of the user.

#ifndef STRUCTURES_DEFINED 
#define STRUCTURES_DEFINED 

#include <vector>
#include <assert.h>
#include "constants.h"
#include "UPTesting.h"

using namespace std;

//format string for printing large integers
#ifdef _MSC_VER
typedef __int64  int64;
#define i64d "I64d"
#else
typedef long long int64;
#define i64d "lld"
#endif

typedef int clause_id;          // < 0 for original, > 0 for learned, 0 reserved
typedef int atom;               // 1-based
typedef int literal;            // low bit is sign (0 is true)

inline bool    negated(literal l)  { return l & 1; }
inline literal negation(literal l) { return l ^ 1; }
inline atom    get_atom(literal l) { return l >> 1; }

inline literal plit(atom a)        { return a << 1; }
inline literal nlit(atom a)        { return (a << 1) | 1; }

inline size_t  index(literal l)    { return ((l & -2) - (negated(l)? 1 : 2)); }

// converstion to/from usual integer representation (-x for negation)

inline int     lit_to_int(literal l) 
{ return negated(l)? -get_atom(l) : get_atom(l); }
inline literal int_to_lit(int x)  { return x < 0? nlit(-x) : plit(x); }
inline atom    int_to_atom(int x) { return x < 0? -x : x; }

/* A variable (as opposed to an atom) is something that we're trying to
value in the problem. */

class variable {
  size_t  level;                // level at which value was set
  literal status;               // current value (0 if unset)
  int     reason;               // why it has this value (0 if branch choice)
  double  score;                // importance of branching on me
  literal saved;                // saved value from previous subtrees

  public:
  variable() : level(0), status(0), reason(0), score(0), saved(0) { }

  friend class heap;
  friend class solver;
};

/* Priority queue where priority can change on the fly */

class heap : public vector<atom> {
  vector<int>             location; // where is this atom in the heap?
                                    // -1 if unplaced
  const vector<variable> *vars;     // pointer; needed for comparison function

  bool cmp(atom i, atom j) const { return (*vars)[i].score > (*vars)[j].score; }
  void move_up(int idx);
  void move_down(int idx);

  public:
  heap() : vars(NULL) { }
  heap(const vector<variable> &v) : location(v.size(),-1), vars(&v) { }

  void clear();
  void insert(atom v);
  void update(atom v) { if (contains(v)) move_up(location[v]); }
                                // assumes score got bigger!
  atom front();                 // pops off queue as well
  bool contains(atom v) const { return location[v] >= 0; }
  void insert_if_missing(atom v) { if (!contains(v)) insert(v); }
};

struct clause : vector<literal> {
  clause_id id;

  clause(int n = 0) : vector<literal>(n) { }
};

struct statistics {
  int64  decisions;               // number made
  int64  conflicts;               // number found
  int    kb_reduction_count;      // how often we deleted bad nogoods
  int    kb_simplification_count; // how often we did it

  statistics() : decisions(0), conflicts(0), kb_reduction_count(0),
    kb_simplification_count(0) { }
};

struct complete_statistics : statistics {
  mutable clock_t start_time;
  int64 clauses;                // total number ever learned
  int64 learned_lits;           // current number of lits (total)
  int64 orig_lits;              // current number of learned lits

  mutable statistics previous;

  complete_statistics() : clauses(0), learned_lits(0), orig_lits(0) { }
};

enum printmode { RESULT, SILENT, SOLUTION };
enum Status { SOLVING = -2, UNKNOWN, UNSATISFIABLE, SATISFIABLE }; 

/* Small data structure to indicate whether or not we should use a
saved value for a variable.  This goes in phases -- use, don't use,
etc.  save_progress indicates whether we're currently in a "use it"
phase and the rest of the stuff is about when to switch.  */

class saving {
  bool   save_progress;        // is it on or off?
  size_t on_th;                // how long to turn on
  size_t on_th_inc;            // increment to on_th
  size_t off_th;               // how long to turn off
  size_t off_th_inc;           // increment
  size_t next_switch;          // next state switch

  public:
  saving() : save_progress(false), on_th(ON_TH), on_th_inc(ON_TH_INC),
    off_th(OFF_TH), off_th_inc(OFF_TH_INC), next_switch(ON_TH) { }

  bool save() { return save_progress; }
  void flip_status();
  void check_status(size_t ct) { if (ct >= next_switch) flip_status(); }
};

struct restart {
  int64   next;
  size_t  number;
  size_t  luby_unit;

  restart() : next(LUBY_UNIT), number(0), luby_unit(LUBY_UNIT) { }
};

/* Solver instance.  An instance of complete_statistics, just to keep
access to the statistical information simple.  */

class solver : complete_statistics {
  vector<clause>    original_clauses; // index -1, -2, ...
  vector<clause>    learned_clauses;  // index 1, 2, ...
  vector<double>    learned_scores;
  vector<variable>  variables;

  public:
  clause &operator[](clause_id i) 
  { return (i < 0)? original_clauses[-1 - i] : learned_clauses[i - 1]; }
  const clause &operator[](clause_id i) const
  { return (i < 0)? original_clauses[-1 - i] : learned_clauses[i - 1]; }

  size_t  level (atom i) const { return variables[i].level; }
  literal status(atom i) const { return variables[i].status; }
  int     reason(atom i) const { return variables[i].reason; }
  double  score (atom i) const { return variables[i].score; }
  literal saved (atom i) const { return variables[i].saved; }

  size_t  & level (atom i) { return variables[i].level; }
  literal & status(atom i) { return variables[i].status; }
  int     & reason(atom i) { return variables[i].reason; }
  double  & score (atom i) { return variables[i].score; }
  literal & saved (atom i) { return variables[i].saved; }

  literal lit_status(literal l) const { return status(get_atom(l)); }
  bool    is_set    (literal l) const { return lit_status(l) == l; }
  bool    is_free   (literal l) const { return lit_status(l) == 0; }
  bool    is_false  (literal l) const { return lit_status(l) == negation(l); }

  bool    empty     ()          const { return original_clauses.empty(); }

  private:
  vector<literal>      assign;        // current partial assignment stack
  vector<int>          decision_lit;  // [n] = # of lits valued before branch #n
  vector<vector<clause_id> > watched; // watched literals for lit -1,1,-2,2,etc

  restart r;                   // restart info
  heap    units;               // unit queue
  saving  save;                // progress saving

  // variable ordering
  heap var_order_heap;         // queue itself
  double score_inc;            // current variable score increment
  
  // learned clause ordering
  double clause_score_inc;     // current clause score increment
  double max_learned_clauses;  // maximum number of learned clauses allowed
  
  // database simplification
  size_t num_conflicts_for_next_simplify; // number of conflicts when to do it
  bool   simplify_orig_kb;
  bool   simplify_learned_kb;

  // printing
  printmode     pmode;
  bool          vbse;
  
  public:
  solver(const vector<set<UPTestingLit> >& clauses, bool &ok);
  void resize(size_t n);

  printmode   print_mode() const { return pmode; }
  printmode & print_mode()       { return pmode; }

  bool      verbose()      const { return vbse; }
  bool    & verbose()            { return vbse; }

  void enqueue(literal l);
  void new_orig_clause(const vector<int> &literals);
  bool finish_initialization();
  void backtrack(size_t d = 2);
  Status solve();

  void print_progress_header() const;
  void print_progress() const;
  void print_progress_footer() const;
  void print_clause(const clause &c) const;
  void print_stats() const;
  void print_solution() const;

  private:

  void remove_clause(const clause &c);
  bool is_satisfied(const clause &c) const;

  void simplify(bool learned = true);
  bool is_reason(const clause &c) const;
  void reduce();

  void rescale_variable_scores();
  void rescale_clause_scores();
  void increment_literal_score(literal l);
  void increment_clause_score(clause_id c);

  void lit_watches_clause(literal l, const clause &c) {
    watched[index(l)].push_back(c.id); 
  }
  void lit_unwatches_clause(literal l, const clause &c);
  void initialize_watchers();
  void initialize_watchers(const clause &c);

  inline bool removable(literal l, vector<bool> &levs, vector<bool> &seen) const;
  clause derive_conflict(clause_id conf_id, size_t clevel, size_t &alevel);

  void add_conflict(clause &conf, size_t lev);

  clause_id process_unit_literal_queue();
  clause_id bcp(literal l, bool push = true);
  clause_id assert_conflict(clause &conf, size_t lev);

  clause_id set_decision(literal l);
  atom first_unvalued() const;
  atom select_variable();

  void correct_watched_lists(bool orig);

  bool timed_out() const;
  Status terminate(Status r);
  void do_restart();
  void do_solved();

  friend struct sortable;
};

// function for printing out a comment line
#define rprintf(format, args...) ( printf("c "), printf(format , ## args), fflush(stdout) )

// time measuring function
#ifdef _MSC_VER
#include <ctime>
static inline double get_cpu_time(void) {
    return (double)clock() / CLOCKS_PER_SEC; }
#else
#include <sys/time.h>
#include <sys/resource.h>
#endif

#endif
